from snowflake.snowpark import Session, Row
from snowflake.connector.errors import ProgrammingError, DatabaseError
import logging
import platform
import os
import toml
import sys

class ConnectionFile:
    def __init__(self, environment: str):
        self.environment=environment
        self.config_path = self.get_config_path()
        logging.info(f'opening {self.config_path}')
        self.config: dict = self.get_config_dict(self.config_path)
        self.required_fields = ["account", "role", "warehouse"]
        self.validate_config_keys()

    def get_config_path(self) -> str:
        system = platform.system() 
        user_root = os.path.expanduser('~')
        default_path = os.path.join(user_root, '.snowflake', 'connections.toml')

        if os.path.exists(default_path):
            return default_path
        elif system == "Windows":
            return self.get_windows_config_path()
        elif system == "Linux":
            return self.get_linux_config_path()
        elif system == "Darwin":
            return self.get_mac_config_path()
        else:
            logging.error(f"Operating system value not found. Expected values of Windows,Linux, or Darwin for platform.system() Found:"+system)
            raise
        
    def get_windows_config_path(self) -> str:
        try:
            base = os.path.expandvars('%USERPROFILE%')
            return os.path.join(base,'AppData','Local','snowflake', 'connections.toml')
        except Exception as e:
            logging.error(f"Unexpected error when reading Snowflake config file on Windows OS: {e}")
            raise
    
    def get_linux_config_path(self) -> str:
        try:
            base_path= os.path.expanduser('~')
            return os.path.join(base_path,'.config','snowflake','connections.toml')
        except Exception as e:
            logging.error(f"Unexpected error when reading Snowflake config file on Linux OS: {e}")
            raise

    def get_mac_config_path(self) -> str:
        try:
            base_path= os.path.expanduser('~')
            return os.path.join(base_path,'Library','Application Support','snowflake','connections.toml')
        except Exception as e:
            logging.error(f"Unexpected error when reading Snowflake config file on Linux OS: {e}")
            raise

    def get_config_dict(self, toml_path) -> dict:
        try:
            with open(toml_path, "r") as toml_file:
                config = toml.load(toml_file)
                return config[self.environment]
        except toml.TomlDecodeError as tde:
            raise  toml.TomlDecodeError(f'Error parsing the connections.toml file: {tde}')
        except FileNotFoundError:
            raise FileNotFoundError(f"Could not find connections.toml file at {toml_path}.")
        except Exception as e:
            raise Exception(f"An error occurred while reading the TOML file: {e}")
        except KeyError:
            raise KeyError(f"Could not find environment key {self.environment} in connections.toml file")

    def validate_config_keys(self) -> bool:
        if not set(self.required_fields) <= set(self.config.keys()):
            logging.error(f'Could not find all required connection parameters in environment {self.environment} in connection file {self.config_path}')
            raise

class SnowflakeUser:
    def __init__(self, environment: str):
        if not environment:
            raise ValueError("Environment not specified. Please provide a valid environment.")
        self.environment = environment
        self.connection_file = ConnectionFile(self.environment)
        self.connection_config = self.connection_file.config
        self.session = self._get_session()
        
    def _connect_with_rsa(self):
        try:
            private_key_path = self.connection_config.get("private_key_path")

            if not os.path.isabs(private_key_path):
                private_key_path = os.path.join(os.getcwd(), private_key_path)

            if private_key_path:
                with open(private_key_path, "rb") as key_file:
                    private_key = key_file.read()
                self.connection_config['private_key']= private_key
                self.connection_config['authenticator']= "snowflake_jwt"
                session_builder = Session.builder.configs(self.connection_config)
                return session_builder.create()
            else:
                raise ValueError(f"Private key not found for environment '{self.environment}' in the TOML file. Looked in path {private_key_path}")
        except FileNotFoundError as e:
            logging.error(f"Private key file not found at path: {private_key_path}. Error: {e}")
            raise
        except (ProgrammingError, DatabaseError) as db_error:
            logging.error(f"Snowflake connection error: {db_error}")
            raise
        except Exception as e:
            logging.error(f"Unexpected error occured during RSA authentication: {e}")
            raise

    def _connect_with_password(self):
        try:
            session = Session.builder.config("connection_name", self.environment).create()
            return session
        except (ProgrammingError, DatabaseError) as db_error:
            logging.error(f"Snowflake connection error: {db_error}")
            raise
        except Exception as e:
            logging.error(f"Unexpected error occured during connection: {e}")
            raise

    def _connect_with_sso(self):
        try:
            session = Session.builder.config("connection_name", self.environment).create()

            return session

        except (ProgrammingError, DatabaseError) as db_error:
            logging.error(f"Snowflake connection error: {db_error}")
            raise
        except Exception as e:
            logging.error(f"Unexpected error occurred during SSO authentication: {e}")
            raise

    def _get_session(self) -> Session:
        try:
            if self.connection_config.get("private_key_path"):
                return self._connect_with_rsa()
            elif self.connection_config.get("authenticator") == "externalbrowser":
                return self._connect_with_sso()
            elif self.connection_config.get("user") and self.connection_config.get("password"):
                return self._connect_with_password()
            else:
                raise ValueError(f"Missing login credentials. Please add either 'private_key_path' for RSA key-pair authentication or 'user' and 'password' for username-password authentication, or authenticator: externalbrowser for SSO authentication")
        except ValueError as ve:
            logging.error(f"ValueError: {ve}")
            raise
        except (ProgrammingError, DatabaseError) as db_error:
            logging.error(f"Snowflake connection error: {db_error}")
            raise
        except Exception as e:
            logging.error(f"Unexpected error: {e}")
            raise RuntimeError('Failed to establish Snowflake session')

    def run_query(self, query:str) -> list[Row]:
        try:
            if query.strip() != '':
                res = self.session.sql(query).collect()
            else:
                res = [Row()]    
            return res
        except (ProgrammingError, DatabaseError) as db_error:
            logging.error(f"Snowflake query execution error: {db_error}")
            logging.error(f"Query: {query}")
            raise
        except Exception as e:
            logging.error(f"Error during query execution: {e}. Query: {query}. Skipping query execution.")
    
    def run_queries(self, queries: list, object_type: str = "") -> list:
        """
        Executes a list of queries in order and returns a list of output results.
        Logs and skips execution if no queries are provided.
        """
        outp = []

        for index, query in enumerate(queries):
            try:
                logging.info(f"Executing query {index + 1}/{len(queries)} for {object_type}")
                logging.debug(f"Executing query {index + 1}/{len(queries)} for {object_type}: {query}")
                result = self.run_query(query)
                outp.append(result)
                logging.debug(f"Query executed successfully: {result}")
            except (ProgrammingError, DatabaseError) as db_error:
                logging.error(f"Database error in {object_type} query execution: {db_error}")
                logging.error(f"Query: {query}")
                raise
            except Exception as e:
                logging.error(f"Unexpected error in {object_type} query execution: {e}. Query: {query}. Skipping query.")
    
        logging.debug(f"Completed all queries for {object_type}, with {len(outp)} successful executions.")
        return outp
    
    def post_files(self, file_configs: list[dict]) -> list:
        outp = []
        for file_config in file_configs:
            logging.info(f"Loading local file {file_config['local_path']} to {file_config['stage_path']}")
            outp.append(self.session.file.put(file_config['local_path'], file_config['stage_path'], auto_compress=False, overwrite=True))
        return outp
    
if __name__ == "__main__":
    logging.basicConfig(stream=sys.stdout, level=logging.INFO,
                        format='%(asctime)s - %(levelname)s - %(filename)s - %(funcName)s - %(message)s')


import logging
import sys
import argparse
from . import commands

class ArgHandler:
    def __init__(self) -> None:
        self.mapper = self._get_mapper()
        self.parser = self._get_parser()
    
    def _get_mapper(self) -> dict:
        # Map Command Names to their classes
        mapper = {
            'deploy': commands.Deploy,
            'init': commands.Init,
            'clone': commands.Clone,
            'run_script': commands.RunScript,
            'test_dag': commands.TestDAG,
        }
        return mapper

    def _get_parser(self) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(
        description=(
                'SnowFlow Deployment Tool\n'
                'SnowFlow is a command-line tool that simplifies and automates deployments, task management, and script execution within Snowflake environments.\n\n'
                'Usage:\n'
                '  - For general help, use "snowflow -h"\n'
                '  - For command-specific help, use "snowflow <command> -h"\n\n'
            ),
        usage=argparse.SUPPRESS,
        add_help=True,
        formatter_class=argparse.RawTextHelpFormatter
    )
        
        subparsers = parser.add_subparsers(title='Available Commands', metavar='', dest='cmd')
        
        for command_name, command_class in self.mapper.items():
            self.add_command_parser(subparsers, command_name, command_class)
        
        return parser

    def add_command_parser(self, subparsers: argparse._SubParsersAction, command_name: str, command_class: type) -> None:
        command_help = command_class.help
        command_args = command_class.get_args()
    
        parser = subparsers.add_parser(
            command_name,
            help=command_help,
            description=command_help,
            usage=argparse.SUPPRESS  
        )
    
        if command_name != 'init':
            parser.add_argument('-e', dest='environment', required=True, help='Specify the environment', metavar='')

        for arg in command_args:
            parser.add_argument(arg.option, required=arg.required, help=arg.help, metavar='')

    def exec(self):
        parsed_args = vars(self.parser.parse_args())
        cmd = parsed_args.pop('cmd', None)
        
        if not cmd:
            self.parser.print_help()
            sys.exit(1)
    
        command_class = self.mapper[cmd]
        if cmd == 'init':
            command_instance = command_class()
        else:
            environment = parsed_args.pop('environment', None)
            if not environment:
                raise ValueError("The '-e' argument is required for this command")
        
            command_instance = command_class(environment)

        command_instance.run(parsed_args)

def main():
    logging.basicConfig(stream=sys.stdout, level=logging.INFO,
                        format='%(asctime)s - %(levelname)s - %(filename)s - %(funcName)s - %(message)s')

    handler = ArgHandler()
    handler.exec()

if __name__ == "__main__":
    main()


from pathlib import Path
import yaml
import logging
import networkx as nx
import sys
import os
from snowflow import sql_templates

class ScriptParser:
    def __init__(self, substitutions=dict()):
        self.substitutions: dict = substitutions or {}

    def parse_yaml_file(self, file_path: Path) -> dict:
        '''
        Load a yaml file as a python dictionary
        '''
        try:
            with open(file_path, 'r') as raw_yaml:
                string_file = raw_yaml.read()
                string_file = string_file.replace('\t','  ')
                string_file = self.substitute_vars(string_file)
                data = yaml.safe_load(string_file)
                if data is None:
                    return {}
                return data
        except yaml.YAMLError as e:
            logging.error(f"YAML parsing error in {file_path}. Error: {e}")
            return {}
        except Exception as e:
            logging.error(f"Unexpected error while parsing YAML file: {e}")
            return {}

    def get_path_yamls(self, path: Path) -> list[dict]:
        try:
            yamls=[]
            for f in path.glob("*.yaml"):
                yamls.append(self.parse_yaml_file(f))
            return yamls
        except FileNotFoundError as e:
            logging.error(f"YAML directory not found: {e}")
            raise
        except Exception as e:
            logging.warning(f"Error while retrieving YAML files: {e}. Skipping.")
            return []

    def clean_query(self, query: str) -> str:
        '''
        Given a SQL query, substitute variables and clean up invalid characters
        '''
        query = self.strip_special_chars(query)
        query = self.substitute_vars(query)
        return query

    def substitute_vars(self, query: str) -> str:
        try:
            if self.substitutions:
                for var, val in self.substitutions.items():
                    query = query.replace(var, str(val))
            return query
        except TypeError as te:
            logging.error(f"TypeError in substitute_vars: {te}. Substitutions: {self.substitutions}")
            raise
        except Exception as e:
            logging.error(f"Unexpected error in substitue_vars: {e}")
            logging.error(f"Query: {query}, Substitutions: {self.substitutions}")
            raise
    
    def strip_special_chars(self, query: str) -> str:
        try:
            return query.rstrip()
        except Exception as e:
            logging.error(f"Unexpected error in strip_special_chars: {e}")
            raise

    def clean_query_list(self, queries: list) -> list[str]:
        try:
            cleaned_list = []
            for query in queries:
                cleaned = self.clean_query(query)
                if cleaned!='':
                    cleaned_list.append(cleaned)
            return cleaned_list
        except Exception as e:
            logging.error(e)
            return queries

    def read_clean_file(self, path: Path) -> str:
        return self.clean_query(self.read_file(path))

    def read_file(self, path: Path) -> str:
        """
        Reads a file, checking the current working directory first.
        """
        # Check in the current working directory first
        user_path = os.path.join(os.getcwd(), path)
    
        if os.path.exists(user_path):
            try:
                with open(user_path, 'r') as file:
                    content = file.read()
                return content
            except Exception as e:
                logging.error(f"Error reading file {user_path}: {e}")
                return ''
        else:
            logging.error(f"File not found: {user_path}. Ensure the file exists.")
            return ''

    
    def read_file_queries(self, path: Path, single_transaction=False) -> list[str]:
        '''
        Read a sql file that can contain multiple queries
        '''
        try:
            queries= []
            if os.path.exists(path):
                if single_transaction:
                    queries = [self.read_clean_file(path)]
                else:
                    queries = self.read_clean_file(path).strip(';').split(';')
            return queries
        except Exception as e:
            logging.error(e)
            return []

    def get_path_queries(self, path: Path, single_transaction: bool = False) -> list[str]:
        """
        Loop over path, return contents of files in a list of queries.
        Looks in the current working directory
        """
        user_path = os.path.join(os.getcwd(), path)
        queries = []
        for f in sorted(Path(user_path).glob("*.sql")):
            if single_transaction:
                queries.append(self.read_clean_file(f))
            else:
                queries.extend(self.read_file_queries(f))
    
        logging.debug(queries)
        return queries

class SQLTemplates:
    def __init__(self):
        self.sf_task_template = self.get_template('sf_task_template.sql')
        self.sql_procedure_template = self.get_template('sql_procedure_template.sql')
        self.user_task_template = self.get_template('user_task_template.sql')

    def get_template(self, name):
        p = Path(__file__).with_name(name)
        try:
            with p.open('r') as f:
                return f.read()
        except FileNotFoundError as e:
            logging.error(e)
            logging.error('Could not find sql template')
            sys.exit(1)

class DirectoryHandler:
    def __init__(self):
        self.root_dir = os.getcwd()

    def mkdir(self, path: Path) -> Path:
        try:
            logging.info('Making the directory if not exists: ' + str(path))
            path.mkdir(parents=False, exist_ok=True)
            return path
        except FileNotFoundError as e:
            logging.error(e)
            logging.error('Could not find parent directory')
            sys.exit(1)

    def get_path_lookup(self, root: Path, child_list: list) -> dict:
        #given a list of child objects, create a lookup dictionary
        return {c: Path(root, c) for c in child_list}

    def initialize_directory(self, root: Path, child_lookup: dict) -> dict:
        #create root and child directories
        self.mkdir(root)
        for name, path in child_lookup.items():
            self.mkdir(path)
        Path(root, 'init.sql').touch()
        Path(root, 'grants.sql').touch()
        
        project_root = Path(self.root_dir) 
        query_vars_file = project_root / 'query_variables.yaml'

        if not query_vars_file.exists():
            try:
                query_vars_file.touch()
                logging.info("Making query_variables.yaml if not exists: " + str(project_root))
            except Exception as e:
                logging.error(f"Failed to create query_variables.yaml: {e}")

        return True

    def get_absolute_path(self, relative_path: str) -> Path:
        """
        Given a path relative to the root as a string, return a Path object based on the current working directory.
        """
        return Path(os.path.join(os.getcwd(), 'snowflake', relative_path))

class Environment:
    def __init__(self, environment: str):
        if not environment:
            raise ValueError("Environment not specified. Please provide a valid environment.")
        self.dh = DirectoryHandler()
        self.sp = ScriptParser()
        self.query_variables_file = 'query_variables.yaml'
        self.env_var = environment  
        self.query_variables = self.get_query_variables(self.env_var) 
        self.sp.substitutions = self.query_variables or {}

    def get_query_variables(self, env):
        try:
            local_path = os.path.join(os.getcwd(), 'query_variables.yaml')
            logging.debug(f"Looking for query variables in: {local_path}")

            if not os.path.exists(local_path):
                logging.debug("query_variables.yaml not found.")
                return {}

            raw_vars = self.sp.parse_yaml_file(local_path)

            if raw_vars is None:
                logging.debug("query_variables.yaml is empty.")
                return {}

            if env not in raw_vars:
                logging.debug(f"No variables found for environment '{env}' in query_variables.yaml.")
                return {}

            logging.info(f"Query variables found for environment '{env}': {raw_vars[env]}")
            return raw_vars[env]
        
        except Exception as e:
            logging.error(f"Unexpected error in get_query_variables: {e}")
            return {}


class SnowflakeAcct:
    '''
    Provides an interface between the repo files and a Snowflake account 
    '''
    def __init__(self, environment: str = None) -> None:
        if environment:
            self.environment = Environment(environment)
            self.sp = self.environment.sp
            self.dh = self.environment.dh
        else:
            self.environment = None
            self.sp = ScriptParser()
            self.dh = DirectoryHandler() 
        
        self.env_dir = Path(os.getcwd(), 'snowflake')
        self.child_objects = ['databases', 'integrations', 'roles', 'warehouses', 'network_rules', 'network_policies']
        self.child_lookup = self.dh.get_path_lookup(self.env_dir, self.child_objects)

    def initialize(self) -> None:
        Path(self.env_dir, 'init.sql').touch()
        Path(self.env_dir, 'grants.sql').touch()

    def get_path_objects(self, obj_type:str, single_transaction: bool=False) -> list[str]:
        '''
        Look up scripts based on string type (roles, tables, views, etc)
        Use single transaction is a file shouldn't be split into multiple transactions by semicolons
        '''
        return self.sp.get_path_queries(self.child_lookup[obj_type],single_transaction)

    def get_grants(self) -> list[str]:
        return self.sp.read_file_queries(Path(self.env_dir, 'grants.sql'))
    
    def get_init(self) -> list[str]:
        return self.sp.read_file_queries(Path(self.env_dir, 'init.sql'))
    
    def get_databases(self) -> list:
        #return all the database objects
        return [SnowflakeDB(d.stem,self) for d in self.child_lookup['databases'].iterdir() if d.is_dir()]

class SnowflakeDB:
    def __init__(self, name: str, account: SnowflakeAcct):
        self.account=account
        self.sp = account.sp
        self.name=name
        self.dh = account.dh

        self.path = Path(os.getcwd(), 'snowflake', 'databases', self.name)
        self.child_objects= ['schemas', 'dml']
        self.path_lookup = self.dh.get_path_lookup(self.path, self.child_objects)
        self.dml_path = Path(self.path, 'dml')
        self.grants_files = Path(self.path, 'grants.sql')
        self.init_file = Path(self.path, 'init.sql')
    
    def __str__(self):
        return self.name
    
    def __repr__(self):
        return self.name
    
    def get_db_init(self):
        return self.sp.read_file_queries(self.init_file)
    
    def get_grants(self):
        return self.sp.get_path_queries(self.grants_files)
    
    def get_schemas(self):
        return [SnowflakeSchema(s.stem,self) for s in self.path_lookup['schemas'].iterdir() if s.is_dir()]

    def initialize(self):
        self.account.initialize()
        self.dh.initialize_directory(self.path, self.path_lookup)
        return True

class SnowflakeSchema:
    def __init__(self, name: str, database: SnowflakeDB):
        self.name = name
        self.database = database
        self.account= self.database.account
        self.environment = self.account.environment
        self.query_variables = self.environment.query_variables
        self.sp = database.sp
        self.sp.substitutions = self.sp.substitutions | self.query_variables
        self.dh = database.dh
        self.schema_path = Path(os.getcwd(), 'snowflake', 'databases', self.database.name, 'schemas', self.name)
        self.child_objects= ['file_formats', 'tables','streams','stages','views','tasks','dags','udfs', 'stored_procs', 'staged_files', 'post_deploy']
        self.path_lookup = self.dh.get_path_lookup(self.schema_path, self.child_objects)
    
    def __str__(self):
        return self.database.name+'.'+self.name
    
    def __repr__(self):
        return self.database.name+'.'+self.name
    
    def get_schema_init(self) -> list[str]:
        return self.sp.read_file_queries(Path(self.schema_path,'init.sql'))

    def get_schema_grants(self) -> list[str]:
        return self.sp.read_file_queries(Path(self.schema_path,'grants.sql'))

    def initialize(self):
        self.account.initialize()
        self.database.initialize()
        self.dh.initialize_directory(self.schema_path, self.path_lookup)
        return True
    
    def get_path_objects(self, obj_type:str, single_transaction: bool=False) -> list[str]:
        '''
        Look up scripts based on string type (roles, tables, views, etc)
        Use single transaction is a file shouldn't be split into multiple transactions by semicolons
        '''
        return self.sp.get_path_queries(self.path_lookup[obj_type],single_transaction)

    def get_tables(self) -> list[str]:
        return self.sp.get_path_queries(self.path_lookup.get('tables'))
    
    def get_stored_procs(self) -> list[str]:
        return self.sp.get_path_queries(self.path_lookup.get('stored_procs'), single_transaction=True) if self.path_lookup.get('stored_procs') else []
    
    def get_grants(self) -> list[str]:
        return self.sp.read_file_queries(Path(self.schema_path,'grants.sql')) 
    
    def get_dags(self) -> list[str]:
        queries = []
        dag_list = [TaskDAG(cd, self) for cd in self.sp.get_path_yamls(self.path_lookup['dags'])]
        for dag in dag_list: 
            queries.extend(dag.get_all_queries())
        return queries

    def get_dag_objs(self):
        return [TaskDAG(cd, self) for cd in self.sp.get_path_yamls(self.path_lookup['dags'])]
    
    def get_dag(self, file_name) -> list[str]:
        curr_dag_path = Path(self.path_lookup['dags'], file_name) if self.path_lookup.get('dags') else None
        if curr_dag_path:
            logging.debug(curr_dag_path)
            cd = self.sp.parse_yaml_file(curr_dag_path)
            dag = TaskDAG(cd, self)
            return dag.get_all_queries()
        else:
            logging.debug(f"Skipping dag {file_name}, folder does not exist.")
            return []
    
    def get_staged_files(self) -> list[dict]:
        """
        Return a list of dicts with the format {'local_path': ,'stage_path':}
        where stage is a stage name that corresponds to a subfolder name in the staged_files
        and path is the filename and path within the folder.
        """
        outp = []
        staged_files_path = Path(self.schema_path, 'staged_files')
        if staged_files_path.exists():
            stages = [p for p in list(staged_files_path.glob('*')) if p.is_dir()]
            for stage in stages:
                local_stage = Path(staged_files_path, stage)
                files = [f for f in list(stage.glob('**/*')) if f.is_file()]
                for file in files:
                    local_path = str(file)
                    stage_path = '@' + stage.name + '/' + file.relative_to(local_stage).parent.as_posix()
                    outp.append({'local_path': local_path, 'stage_path': stage_path})
        else:
            logging.debug(f"Skipping staged files, folder does not exist.")
        return outp

class TaskDAG:
    def __init__(self, config_dict: dict, schema: SnowflakeSchema):
        self.config_dict=config_dict
        logging.info(self.config_dict)
        self.sql_templates  = SQLTemplates()
        self.schema= schema
        self.name = self.config_dict.get('DAG_NAME')
        self.root= self.config_dict.get('ROOT_TASK')
        self.sp = schema.sp
        self.dh = schema.dh
        self.query_variables = self.get_query_variables()
        self.task_template: str = self._get_task_template()
        self.sp.substitutions = self.query_variables
        self.task_dict, self.digraph = self._get_structs()
         
        
    
    def _get_task_dict(self) -> dict:
        #check for duplicate names
        #check for required value
        task_dict = dict()
        for t_config in self.config_dict['TASKS']:
            task = Task(t_config)
            task_dict[task.name]=task
        return task_dict
    
    def _get_structs(self):
        digraph  = nx.DiGraph()
        task_dict = dict()
        for t_config in self.config_dict['TASKS']:
            curr_name = t_config['NAME']
            dependencies = t_config.get('DEPENDS_ON',[])
            digraph.add_node(curr_name)
            for dependency in dependencies:
                digraph.add_edge(dependency,curr_name)

            task = Task(t_config, self)
            task_dict[task.name]=task
        return task_dict, digraph

    def get_all_queries(self):
        query_list=[]
        create_order = list(nx.topological_sort(self.digraph))
        for task in create_order:
            t = self.task_dict[task]
            query_list.extend(t.get_create_queries())
        if self.config_dict.get('ENABLED')=='TRUE':
            dependents =  "SELECT SYSTEM$TASK_DEPENDENTS_ENABLE('"+self.root+"')"
            root_enable = "ALTER TASK "+self.root+" RESUME"
            query_list.append(dependents)
            query_list.append(root_enable)
        else:
            logging.info('suspend task')
            root_disable = "ALTER TASK "+self.root+" SUSPEND"
            query_list.append(root_disable)
        return query_list
    
    def get_query_variables(self) -> dict:
        var_dict = self.schema.query_variables
        keys = ['ROOT_TASK','INITIAL_WAREHOUSE_SIZE','ALLOW_OVERLAPPING_EXECUTION','WAREHOUSE']
        for key in keys:
            var_dict['!!!'+key+'!!!'] = self.config_dict.get(key)
        logging.debug(var_dict)
        return var_dict
    
    def _get_task_template(self) -> str:
        '''
        Returns the file name of the template to use
        '''
        template = 'NO_TEMPLATE_FOUND'
        if 'INITIAL_WAREHOUSE_SIZE' in self.config_dict.keys():
            template = self.sql_templates.sf_task_template 
        elif 'WAREHOUSE' in self.config_dict.keys():
            template = self.sql_templates.user_task_template
        elif 'INITIAL_WAREHOUSE_SIZE' in self.config_dict.keys() and 'WAREHOUSE' in self.config_dict.keys():
            logging.error('Could not determine template to use. Please specify a variable of either INITIAL_WAREHOUSE_SIZE or WAREHOUSE')
        else:
            logging.error('Could not determine template to use. Please specify a variable of either INITIAL_WAREHOUSE_SIZE or WAREHOUSE')
        logging.debug(template)
        return template

class Task:
    def __init__(self, config_dict: dict, dag: TaskDAG):
        self.config_dict = config_dict
        self.name = self.config_dict.get('NAME')
        self.dag= dag
        self.schema = dag.schema
        self.database= self.schema.database
        self.account= self.database.account
        self.is_root = self._is_root()
        self.script_path = Path(self.database.dml_path, self.config_dict.get('SCRIPT_PATH'))
        self.sp = ScriptParser(self.dag.sp.substitutions.copy())
        self.sp.substitutions.update(self.get_query_variables())
    
    def get_create_queries(self) -> list[str]:
        l = list()
        if self.is_root:
            l.append('ALTER TASK IF EXISTS '+self.name+ ' suspend')
        l.append(self.get_sql_proc_code())
        l.append(self.get_task_code())
        return l

    def _is_root(self) -> bool:
        if self.name == self.dag.root:
            return True
        else:
            return False
    
    def get_sql_proc_code(self) -> list[str]: 
        try:
            sql_proc_code = self.sp.clean_query(self.dag.sql_templates.sql_procedure_template)
            logging.info(f"Generated SQL Procedure for {self.name}")
            logging.debug(f"Generated SQL Procedure for {self.name}: \n{sql_proc_code}")
            return sql_proc_code
        except Exception as e:
            logging.error(e)
    
    def get_when_clause(self) -> str:
        value = self.config_dict.get('WHEN', None)
        if not value:
            return ' '
        else:
            return 'WHEN '+value+' '
    
    def get_query_variables(self) -> dict:
        var_dict=dict()
        var_dict['!!!DAG_NAME!!!']=self.dag.name
        var_dict['!!!TASK_NAME!!!']=self.name
        var_dict['!!!PROC_NAME!!!']=self.name+'_sp'
        var_dict['!!!WHEN_CLAUSE!!!'] = self.get_when_clause()
        var_dict['!!!SCRIPT_CODE!!!'] = self.get_script_code()
        if self.is_root:
            overlap = 'ALLOW_OVERLAPPING_EXECUTION = '+self.dag.config_dict.get('ALLOW_OVERLAPPING_EXECUTION') 
            suspend = 'SUSPEND_TASK_AFTER_NUM_FAILURES = 1'
            var_dict['!!!ROOT_VARIABLES!!!']=overlap+' '+suspend
            var_dict['!!!SCHEDULE!!!']= "SCHEDULE = '"+self.dag.config_dict.get('SCHEDULE','')+"'"
        else:
            var_dict['!!!ROOT_VARIABLES!!!']= ''
            var_dict['!!!SCHEDULE!!!'] = 'AFTER '+','.join(self.config_dict.get('DEPENDS_ON',''))
        return var_dict

    def get_script_code(self) -> str:
        query_list =  self.sp.read_file_queries(self.script_path)
        return ';'.join(query_list)
    
    def get_schedule(self) -> str:
        #if root, "SCHEDULE = 'USING CRON 0 8 * * * America/New_York'"
        if self.is_root:
            schedule = "SCHEDULE = '"+self.dag.config_dict.get('SCHEDULE','')+"'"
        else:
            schedule = 'AFTER '+','.join(self.config_dict.get('DEPENDS_ON',''))
        return schedule
    
    def get_task_code(self) -> str:
        return self.sp.clean_query(self.dag.task_template)
    

if __name__ == "__main__":
    logging.basicConfig(stream=sys.stdout, level=logging.INFO,
                        format='%(asctime)s - %(levelname)s - %(filename)s - %(funcName)s - %(message)s')

    acct =SnowflakeAcct('DEV_ACCOUNT')

    
    db = SnowflakeDB('demo',acct)
    schema = SnowflakeSchema('opendata', db)
    print(schema.get_path_objects('views'))




from setuptools import setup, find_packages

setup(
    name="snowflow",
    version="0.0.5",
    author="Thomas Garcia, Aryan Singh",
    author_email="tgarcia@svam.com, aryan.singh@svam.com",
    description="SnowFlow - A Snowflake Deployment Tool",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "snowflake-connector-python",
        "snowflake-snowpark-python",
        "snowflake-connector-python[secure-local-storage]",
        "networkx",
        "pyyaml",
        "toml",
    ],
    entry_points={
        'console_scripts': [
            'snowflow = snowflow.handler:main', 
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    license="BSD-3-Clause",
    python_requires='>=3.6',
)


import logging
from . import runner
from . import scripts
from snowflake.connector.errors import ProgrammingError, DatabaseError
import sys

class Argument:
    def __init__(self, option, required, help):
        self.option = option
        self.required = required
        self.help = help

class SnowFlowCommand:
    def __init__(self) -> None:
        self.name = None
        self.help = None
        self.args = self.get_args()

    def get_args(self) -> list[Argument]:
        pass

    def run(self, options) -> None:
        pass

class Deploy:
    help = 'Deploy account, database, or schema objects. Requires -e to specify the environment.'
    args = [
        Argument('-d', False, 'Specify Database name - Should match the folder'),
        Argument('-s', False, 'Specify Schema name - Should match the folder')
    ]

    def __init__(self, environment: str = None) -> None:
        self.name = 'Deploy'
        self.environment = environment
        self.args = self.get_args()
    
    @classmethod
    def get_args(cls):
        return cls.args
    
    def run(self, args: dict) -> None:
        if self.environment is None:
            raise ValueError("The '-e' argument is required for 'deploy' command.")
        try:
            user = runner.SnowflakeUser(self.environment)
            if args.get('d')==None:
                logging.info('Snowflow deploy account')
                self.account(user)
            elif args.get('s')==None:
                logging.info('Snowflow deploy db')
                self.database(user, args.get('d'))
            else:
                logging.info('Snowflow deploy schema')
                self.schema(user, args.get('d'), args.get('s'))
        except ValueError as ve:
            logging.error(f"Deployment error: {ve}")
            raise 
        except DatabaseError as de:
            logging.error(f"Database error during deployment: {de}")
            raise
        except Exception as e:
            logging.error(f"Unexpected error during deployment: {e}")
            logging.warning("Continuing execution; non-critical error")

    def account(self, user: runner.SnowflakeUser) -> None:
        user.session.query_tag= 'Snowflow deploy account'
        acct = scripts.SnowflakeAcct(self.environment)
        user.run_queries(acct.get_init(), object_type = "init")

        user.run_queries(acct.get_path_objects('roles'), object_type = "roles")
        user.run_queries(acct.get_path_objects('warehouses'), object_type = "warehouses")
        user.run_queries(acct.get_path_objects('integrations'), object_type = "integrations")
        user.run_queries(acct.get_path_objects('network_rules'), object_type = "network_rules")
        user.run_queries(acct.get_path_objects('network_policies'), object_type = "network_policies")
        user.run_queries(acct.get_grants(), object_type = "grants")
        logging.info('Account deployed')

    def database(self, user: runner.SnowflakeUser, db_name:str) -> None:
        user.session.query_tag= 'Snowflow Deploy Database: '+db_name
        acct = scripts.SnowflakeAcct(self.environment)
        db = scripts.SnowflakeDB(db_name,acct)
        user.run_queries(db.get_db_init())
        logging.info('Database Deployed')

    def schema(self, user: runner.SnowflakeUser, db_name:str, schema_name: str) -> None:
        user.session.query_tag= 'Snowflow Deploy Schema: '+db_name+'.'+schema_name
        acct = scripts.SnowflakeAcct(self.environment)
        db = scripts.SnowflakeDB(db_name,acct)
        schema = scripts.SnowflakeSchema(schema_name, db)

        user.run_queries(schema.get_schema_init(),  object_type = "init")
        user.session.use_schema(schema_name)

        user.run_queries(schema.get_path_objects('file_formats'), object_type = "file_formats")
        user.run_queries(schema.get_path_objects('stages'), object_type = "stages")
        user.post_files(schema.get_staged_files())
        user.run_queries(schema.get_path_objects('udfs'), object_type = "udfs")
        user.run_queries(schema.get_path_objects('tables'), object_type = "tables")
        user.run_queries(schema.get_path_objects('views'), object_type = "views")
        user.run_queries(schema.get_path_objects('streams'), object_type = "streams")
        user.run_queries(schema.get_path_objects('stored_procs', single_transaction=True), object_type = "stored_procs")
        user.run_queries(schema.get_path_objects('tasks'), object_type = "tasks")
        user.run_queries(schema.get_dags(), object_type = "dags")
        user.run_queries(schema.get_path_objects('post_deploy'), object_type = "post_deploy")
        user.run_queries(schema.get_grants(), object_type = "grants")
        logging.info('Schema Deployed')

class Init:
    help = 'Initialize folder structure for account, database, or schema. Does not require -e.'
    args = [
        Argument('-d', False, 'Specify Database name for initialization (optional)'),
        Argument('-s', False, 'Specify Schema name for initialization (optional)')
    ]

    def __init__(self) -> None:
        self.name = 'init'

    @classmethod
    def get_args(cls):
        return cls.args
    
    def run(self, args: dict) -> None:
        try:
            db_name = args.get('d')
            schema_name = args.get('s')
    
            if not db_name:
                logging.info('Snowflow initialize account')
                self.account()
            elif not schema_name:
                logging.info('Snowflow initialize database')
                self.database(db_name)
            else:
                logging.info('Snowflow initialize schema')
                self.schema(db_name, schema_name)
        except KeyError as ve:
            logging.error(f"Key error: {ve}. Check if database or schema name is provided correctly.")
            raise
        except Exception as e:
            logging.error(f"Unexpected error during initialization: {e}")
            raise

    def account(self) -> None:
        acct = scripts.SnowflakeAcct()
        logging.info(acct.initialize())

    def database(self, db_name:str) -> None:
        acct = scripts.SnowflakeAcct()
        db = scripts.SnowflakeDB(db_name,acct)
        logging.info(db.initialize())

    def schema(self, db_name:str, schema_name: str) -> None:
        acct = scripts.SnowflakeAcct()
        db = scripts.SnowflakeDB(db_name,acct)
        schema = scripts.SnowflakeSchema(schema_name, db)
        logging.info(schema.initialize())

class Clone:
    help = 'Clone a Database or Schema. Specify just the database to clone the database. Specify schema to clone a schema. Requires -e to specify the environment.'
    args = [
        Argument('-sd', True, 'Source Database name'),
        Argument('-ss', False, 'Source Schema name. Specify to clone a schema'),
        Argument('-td', True, 'Target Database name'),
        Argument('-ts', False, 'Target Schema name. Specify to clone a schema')
    ]

    def __init__(self, environment: str = None) -> None:
        self.name = 'init'
        self.environment = environment

    @classmethod
    def get_args(cls):
        return cls.args
    
    def run(self, args: dict) -> None:
        try:
            ss= args.get('ss')
            ts= args.get('ts')
            sd= args.get('sd')
            td= args.get('td')
        
            if ss==None and ts==None:
                # Clone db
                query = 'create or replace database '+td+ ' clone '+sd
            elif bool(ss)!=bool(ts):
                # If we got one schema instead of two
                logging.error('One schema found. Please submit both source and target schema if cloning schemas')
                query='' 
            else:
                #clone schema
                source = sd+'.'+ss
                tgt = td+'.'+ts
                query = 'create or replace schema '+tgt+' clone '+source
            user = runner.SnowflakeUser(self.environment)
            logging.info(user.run_query(query))

        except ValueError as ve:
            logging.error(f"Clone error: {ve}")
            raise
        except DatabaseError as de:
            logging.error(f"Database error during cloning: {de}")
            raise
        except ProgrammingError as pe:
            logging.error(f"Programming error during cloning: {pe}")
            raise
        except Exception as e:
            logging.error(f"Unexpected error during cloning: {e}")
            raise

class RunScript:
    help = 'Run a specific SQL script. Requires -e to specify the environment.'
    args = [
        Argument('-d', False, 'Database name. Specified in a use_database statement'),
        Argument('-s', False, 'Schema name. Specified in a use_schema statement'),
        Argument('-f', True, 'Script Path relative to the Root folder')
    ]

    def __init__(self, environment: str = None) -> None:
        self.name = 'init'
        self.environment = environment

    @classmethod
    def get_args(cls):
        return cls.args
    
    def run(self, args: dict) -> None:
        try:
            database= args.get('d')
            schema= args.get('s')
            script_path= args.get('f')

            user = runner.SnowflakeUser(self.environment)
            env = scripts.Environment(self.environment)
            path = env.dh.get_absolute_path(script_path)
            queries = env.sp.read_file_queries(path)
            if database:
                user.session.use_database(database)
            if schema:
                user.session.use_schema(schema)
            logging.info(user.run_queries(queries))

        except ValueError as ve:
            logging.error(f"RunScript error: {ve}")
            raise
        except DatabaseError as de:
            logging.error(f"Database error while running script: {de}")
            raise
        except Exception as e:
            logging.warning(f"Error during script execution: {e}. Skipping the script.")

class TestDAG:
    help = 'Test run a DAG file. Requires -e to specify the environment.'
    args = [
        Argument('-d', True, 'Database name'),
        Argument('-s', True, 'Schema name'),
        Argument('-f', True, 'DAG file name relative to DAG folder')
    ]

    def __init__(self, environment: str = None) -> None:
        self.name = 'init'
        self.environment = environment

    @classmethod
    def get_args(cls):
        return cls.args
    
    def run(self, args: dict) -> None:
        database= args.get('d')
        schema_name= args.get('s')
        script_path= args.get('f')
        acct = scripts.SnowflakeAcct(self.environment)
        db = scripts.SnowflakeDB(database,acct)
        schema = scripts.SnowflakeSchema(schema_name, db)
        queries= schema.get_dag(script_path)
        try:
            user = runner.SnowflakeUser(self.environment)
            user.session.use_schema(schema_name)
            logging.info(user.run_queries(queries))
        except DatabaseError as de:
            logging.error(f"Database error during DAG test: {de}")
            raise
        except ProgrammingError as pe:
            logging.error(f"Programming error during DAG test: {pe}")
            raise
        except Exception as e:
            logging.error(f"Unexpected error during DAG test: {e}")
            raise

if __name__ == "__main__":
    logging.basicConfig(stream=sys.stdout, level=logging.INFO,
                        format='%(asctime)s - %(levelname)s - %(filename)s - %(funcName)s - %(message)s')
